<?php declare(strict_types = 1);
/**
 * This file is part of mappi/cursor.
 */

namespace Grifart\Mappi\Cursor;
use Grifart\Mappi\Cursor\Driver\ICursorDriver;

/**
 * PostgreSQL cursor driver
 *
 * Basic envelope over PostgreSQL's cursor. This cursor envelope does not know
 * cursor position, thus cannot do boundary checks. If you reach end, you
 * simply get no more values. You can check if you reached the end by
 * `isHeadOnRecord()`.
 *
 * @package Grifart\Mappi\Cursor
 */
final class Cursor implements ICursor
{
	/** @var ICursorDriver */
	private $driver;

	/** @var string */
	private $name;

	/** @var bool */
	private $headOnRecord = FALSE;

	/**
	 * Warning: this class itself does not check if cursor is really valid in
	 * constructor.
	 *
	 * @param ICursorDriver $driver
	 * @param string        $name Name of cursor (if not valid all later call will fail)
	 */
	public function __construct(ICursorDriver $driver, string $name)
	{
		$this->driver = $driver;
		$this->name = $name;
		// todo: close; for now automatically closed on transaction end; maybe in destructor? Ignore errors?
	}


	public function getName() : string
	{
		return $this->name;
	}

	public function isOnRecord() : bool
	{
		return $this->headOnRecord;
	}

	public function moveTo(int $index)
	{
		if ($index < 0) {
			throw new \InvalidArgumentException("Negative index not supported. Use moveFromEndTo() instead.");
		}
		$this->headOnRecord = $this->driver->moveFromBeginningTo($this->name, $index);
	}

	public function moveFromEndTo(int $index)
	{
		if ($index < 0) {
			throw new \InvalidArgumentException("Negative index not supported. Use moveTo() instead.");
		}

		$this->headOnRecord = $this->driver->moveFromEndTo($this->name, $index);
	}

	public function moveBy(int $offset)
	{
		$this->headOnRecord = $this->driver->moveBy($this->name, $offset);
	}

	public function scroll(int $offset) : int
	{
		$numberOfRows = $this->driver->scroll($this->name, $offset);
		$this->headOnRecord = $numberOfRows === abs($offset);
		return $numberOfRows;
	}

	// ----------------- FETCH ---------------------

	public function fetchRange(int $offset) : array
	{
		$recs = $this->driver->fetchRange($this->name, $offset);

		if ($offset === 0) { // current row
			$this->headOnRecord = count($recs) === 1;
		} else {
			// todo: test this properly!!
			$this->headOnRecord = count($recs) === abs($offset);
		}
		return $recs;
	}

	public function fetchOneAt(int $index)
	{
		$row = $this->driver->fetchOneAt($this->name, $index);
		$this->headOnRecord = $row !== NULL;
		return $row;
	}

	public function fetchOneBy(int $offset)
	{
		$row = $this->driver->fetchOneBy($this->name, $offset);
		$this->headOnRecord = $row !== NULL;
		return $row;
	}

}