<?php
/**
 * This file is part of mappi/cursor.
 */
namespace Grifart\Mappi\Cursor\Driver;

/**
 * Represents cursor
 *
 * Cursor works same as tape. When you have 3 items on it it has these indexes.
 * index:     0  1  2  3  4
 * value:     -  1  2  3  -
 * head:      ^
 *
 * - You have HEAD which you can move around and read records before, after or current.
 * - You cannot ask for current position.
 * - You can move to absolute position from left or from right
 * - You can more relatively to any side
 * - BEGINNING (0) and END (4) has no value
 * - Cursor starts at BEGINNING
 * - When you move too far HEAD will be parked in BEGINNING position or in the END position
 *
 * @package Grifart\Mappi\Cursor
 */
interface ICursorDriver
{
	const FETCH_REMAINING = PHP_INT_MAX;
	const FETCH_FOREGOING = PHP_INT_MIN;

	/**
	 * Move cursor from the BEGINNING to given position
	 * @param string $name The cursor name
	 * @param int    $index
	 * @return bool Head on record?
	 */
	public function moveFromBeginningTo(string $name, int $index) : bool;

	/**
	 * Move cursor from the END to given position
	 * @param string $name The cursor name
	 * @param int    $index
	 * @return bool  Head on record?
	 */
	public function moveFromEndTo(string $name, int $index) : bool;

	/**
	 * Move cursor from the current position by given offset
	 * @param string $name The cursor name
	 * @param int    $offset
	 * @return bool  Head on record?
	 */
	public function moveBy(string $name, int $offset) : bool;

	/**
	 * Scroll from current position by given offset
	 * @param string $name The cursor
	 * @param int    $offset
	 * @return int Number of records scrolled
	 */
	public function scroll(string $name, int $offset) : int;

	/**
	 * Fetch all rows from current position to given offset
	 * @param string $name The cursor name
	 * @param int    $offset
	 * @return array
	 */
	public function fetchRange(string $name, int $offset) : array;

	/**
	 * Fetch one row at given position
	 * @param string $name The cursor name
	 * @param int    $index from the BEGINNING; if negative from the END
	 * @return array|null
	 */
	public function fetchOneAt(string $name, int $index);

	/**
	 * Fetch one row at relative position from current row
	 * @param string $name The cursor name
	 * @param int    $offset
	 * @return array|null
	 */
	public function fetchOneBy(string $name, int $offset);
}