<?php declare(strict_types = 1);
/**
 * This file is part of mappi/cursor.
 */

namespace Grifart\Mappi\Cursor\Driver;

use Dibi\Connection;
use Dibi\Row;

class PostgresCursorDriver implements ICursorDriver
{

	/** @var  Connection */
	private $connection;

	/**
	 * PostgresCursorDriver constructor.
	 * @param Connection $connection
	 */
	public function __construct(Connection $connection)
	{
		$this->connection = $connection;
	}

	/**
	 * @return Connection
	 */
	private function getConnection()
	{
		return $this->connection;
	}

	public function moveFromBeginningTo(string $name, int $index) : bool
	{
		if ($index < 0) {
			throw new \InvalidArgumentException("Negative index not supported. Use moveFromEndTo() instead.");
		}
		return !!$this->getConnection()->query(
			"MOVE ABSOLUTE %i IN %n",
			$index,
			$name
		);
	}

	public function moveFromEndTo(string $name, int $index) : bool
	{
		if ($index < 0) {
			throw new \InvalidArgumentException("Negative index not supported. Use moveTo() instead.");
		}
		if ($index === 0) { // END position
			return !!$this->getConnection()->query(
				"MOVE ABSOLUTE -1 IN %n; MOVE NEXT IN %n;",
				$name,
				$name
			);
		}

		$negativeIndex = -$index;
		return !!$this->getConnection()->query(
			"MOVE ABSOLUTE %i IN %n",
			$negativeIndex,
			$name
		);
	}

	public function moveBy(string $name, int $offset) : bool
	{
		return !!$this->getConnection()->query(
			"MOVE RELATIVE %i IN %n",
			$offset,
			$name
		);
	}

	public function scroll(string $name, int $offset) : int
	{
		if($offset === 0) return 0;
		$forward = $offset > 0;
		$offsetSql = abs($offset);

		if($forward > 0) {
			if($offset === self::FETCH_REMAINING) {
				$offsetSql = "ALL";
			}
			return $this->getConnection()->query(
				"MOVE FORWARD %sql IN %n",
				$offsetSql,
				$name
			);
 		} else {
			if($offset === self::FETCH_FOREGOING) {
				$offsetSql = "ALL";
			}
			return $this->getConnection()->query(
				"MOVE BACKWARD %sql IN %n",
				$offsetSql,
				$name
			);
		}
	}

	// ----------------- FETCH ---------------------

	public function fetchRange(string $name, int $offset) : array
	{
		$forward = $offset >= 0;
		$numberOfRowsSQLClause = ($offset === self::FETCH_FOREGOING || $offset === self::FETCH_REMAINING) ?
			"ALL" : abs($offset);

		$recs = $this->connection->query(
			$forward ? "FETCH FORWARD %sql FROM %n" : "FETCH BACKWARD %sql FROM %n",
			$numberOfRowsSQLClause,
			$name
		)->fetchAll();

		return $this->convertRows($recs);
	}

	public function fetchOneAt(string $name, int $index)
	{
		$row = $this->connection->query(
			"FETCH ABSOLUTE %i FROM %n",
			$index,
			$name
		)->fetch();
		if($row === FALSE) {
			return NULL;
		}
		return $this->convertRow($row);
	}

	public function fetchOneBy(string $name, int $offset)
	{
		$row = $this->connection->query(
			"FETCH RELATIVE %i FROM %n",
			$offset,
			$name
		)->fetch();
		if($row === FALSE) {
			return NULL;
		}
		return $this->convertRow($row);
	}

	// ----------- DATA CONVERSION -----------

	private function convertRow(Row $row) : array
	{
		return $row->toArray();
	}

	/**
	 * @param Row[] $rows
	 * @return array
	 */
	private function convertRows(array $rows) : array
	{
		$new = [];
		foreach ($rows as $row) {
			$new[] = $this->convertRow($row);
		}
		return $new;
	}

}