<?php declare(strict_types = 1);
/**
 * This file is part of mappi/cursor.
 */

namespace Grifart\Mappi\Cursor\Driver;

use Dibi\Connection;
use Grifart\Mappi\Cursor\Cursor;
use Grifart\Mappi\Cursor\ICursor;
use Grifart\Mappi\Cursor\Position;
use Grifart\Mappi\Cursor\SemanticCursor;
use Grifart\Mappi\Cursor\SemanticTrackedCursor;
use Grifart\Mappi\Cursor\TrackedCursor;

final class PostgresCursorFactory
{
	/** @var Connection */
	private $connection;

	/**
	 * PostgresCursorFactory constructor.
	 * @param Connection $connection
	 */
	public function __construct(Connection $connection)
	{
		$this->connection = $connection;
	}

	/**
	 * Creates new instance of Cursor
	 *
	 * Warning: creating cursor requires to be IN transaction block.
	 * Please enter transaction before calling create()
	 *
	 * @param string $sql
	 * @param bool $scroll
	 * @return ICursor
	 */
	public function create(string $sql, bool $scroll) : ICursor
	{
		$this->connection->query(
			"DECLARE %n %SQL CURSOR FOR (%SQL)",
			$name = $this->generateCursorName(),
			$scroll ? "SCROLL" : "NO SCROLL",
			$sql
		);
		
		return new Cursor(
			new PostgresCursorDriver($this->connection),
			$name
		);
	}

	public function createSemantic(string $sql, bool $scroll) : SemanticTrackedCursor
	{
		$cursor = $this->create($sql, $scroll);
		return new SemanticCursor($cursor);
	}

	public function createTrackedSemantic(string $sql, bool $scroll) : SemanticTrackedCursor
	{
		$cursor = $this->create($sql, $scroll);
		return new SemanticTrackedCursor(
			new TrackedCursor($cursor, Position::fromLeft(0))
		);
	}

	/**
	 * Generate cursor unique name in current transaction
	 * @return string
	 */
	private function generateCursorName() : string
	{
		return uniqid();
	}

}