<?php
/**
 * This file is part of mappi/cursor.
 */
namespace Grifart\Mappi\Cursor;

/**
 * Represents cursor
 *
 * Cursor works same as tape. When you have 3 items on it it has these indexes.
 * index:     0  1  2  3  4
 * value:     -  1  2  3  -
 * head:      ^
 *
 * - You have HEAD which you can move around and read records before, after or current.
 * - You cannot ask for current position.
 * - You can move to absolute position from left or from right
 * - You can more relatively to any side
 * - BEGINNING (0) and END (4) has no value
 * - Cursor starts at BEGINNING
 * - When you move too far you end in one of ends
 * 
 * @link http://www.postgresql.org/docs/9.5/static/plpgsql-cursors.html
 * @link http://www.postgresql.org/docs/8.1/static/sql-fetch.html
 * @link http://www.postgresql.org/docs/8.1/static/sql-move.html
 * @package Grifart\Mappi\Cursor
 */
interface ICursor
{
	const FETCH_REMAINING = PHP_INT_MAX;
	const FETCH_FOREGOING = PHP_INT_MIN;

	public function getName() : string;

	public function isOnRecord() : bool;

	/**
	 * Move cursor by offset
	 * @param int $offset
	 * @return void
	 */
	public function moveBy(int $offset);

	/**
	 * Move cursor to absolute position from the beginning
	 * @param int $index
	 * @throws CursorException Negative indexes not supported
	 * @return void
	 */
	public function moveTo(int $index);

	/**
	 * Move cursor to absolute position from the end
	 * @param int $index
	 * @throws CursorException Negative indexes not supported
	 * @return mixed
	 */
	public function moveFromEndTo(int $index);

	/**
	 * Move cursor by given offset and return number of scrolled rows
	 * @param int $offset
	 * @return int The number of scrolled rows
	 */
	public function scroll(int $offset) : int;

	/**
	 * Fetch the next/previous count of rows. If zero given current row is fetched.
	 * Cursor position will be on the last fetched row.
	 * @param int $offset the last row offset
	 * @return array fetched rows
	 */
	public function fetchRange(int $offset) : array;

	/**
	 * Fetch one row at given position
	 * Cursor position will be on the fetched row.
	 * @param int $index Positive means from the beginning; negative from the end
	 * @return array|NULL
	 */
	public function fetchOneAt(int $index);

	/**
	 * Fetch one row at relative position from current position
	 * Has the same effect like calling moveBy() and then fetch current row
	 * Cursor position will be on the fetched row.
	 * @param int $offset
	 * @return array|NULL
	 */
	public function fetchOneBy(int $offset);

}