<?php declare(strict_types = 1);
/**
 * This file is part of mappi/cursor.
 */

namespace Grifart\Mappi\Cursor;

/**
 * Adds semantic API into ICursor.
 *
 * Terminology:
 * - BEGINNING: the position before first row (initial)
 * - FIRST: the first row position
 * - LAST: the last row position
 * - END: the position after last row (end position)
 *
 * @package Grifart\Mappi\Cursor
 */
class SemanticCursor implements ICursor
{
	/** @var ICursor */
	protected $cursor;

	/**
	 * @param ICursor $cursor
	 */
	public function __construct(ICursor $cursor)
	{
		$this->cursor = $cursor;
	}

	/**
	 * Move cursor on the last row
	 * @see fetchCurrent()
	 * @see fetchCurrentSingle()
	 * @return void
	 */
	public function moveToLast()
	{
		$this->moveFromEndTo(1);
	}

	/**
	 * Move cursor on the first row (index=1)
	 * @see fetchCurrent()
	 * @see fetchCurrentSingle()
	 * @return void
	 */
	public function moveToFirst()
	{
		$this->moveTo(1);
	}

	/**
	 * Moves cursor to position BEFORE first row (index=0)
	 * @see fetchNext()
	 * @see fetchNextSingle()
	 * @return void
	 */
	public function moveToBeginning()
	{
		$this->moveTo(0);
	}

	/**
	 * Moves cursor to position AFTER last row
	 * @return void
	 */
	public function moveToEnd()
	{
		$this->moveFromEndTo(0);
	}

	/**
	 * Scrolls to the END
	 */
	public function scrollToEnd() : int
	{
		return $this->scroll(ICursor::FETCH_REMAINING);
	}

	/**
	 * Scroll to the BEGINNING
	 */
	public function scrollToBeginning() : int
	{
		return $this->scroll(ICursor::FETCH_FOREGOING);
	}

	// ---- FETCH ----

	/**
	 * Move cursor to the next position and return row
	 * @return array|NULL the next row
	 * @throws CursorException
	 */
	public function fetchNext()
	{
		return $this->extractOneRowResult(
			$this->fetchRange(1)
		);
	}

	/**
	 * Return row currently standing on
	 * @return array|NULL the current row
	 * @throws CursorException
	 */
	public function fetchCurrent()
	{
		return $this->extractOneRowResult(
			$this->fetchRange(0)
		);
	}

	/**
	 * Move cursor one back and fetch current row
	 * @return array|NULL the previous row
	 * @throws CursorException
	 */
	public function fetchPrevious()
	{
		return $this->extractOneRowResult(
			$this->fetchRange(-1)
		);
	}

	/**
	 * Move cursor to next position and fetch first column value.
	 * @return mixed the first column on next row value
	 * @throws CursorException
	 */
	public function fetchNextSingle()
	{
		return $this->extractFirstColumnFromRow(
			$this->fetchNext()
		);
	}

	/**
	 * Fetch first column value.
	 * @return mixed the first column on current row value
	 * @throws CursorException
	 */
	public function fetchCurrentSingle()
	{
		return $this->extractFirstColumnFromRow(
			$this->fetchCurrent()
		);
	}

	/**
	 * Move cursor to previous position and fetch first column value.
	 * @return mixed the first column on previous row value
	 * @throws CursorException
	 */
	public function fetchPreviousSingle()
	{
		return $this->extractFirstColumnFromRow(
			$this->fetchPrevious()
		);
	}

	/**
	 * Fetch remaining rows. Cursor will be in END position after this operation.
	 * @return array[] remaining rows
	 */
	public function fetchRemaining() : array
	{
		return $this->cursor->fetchRange(
			ICursor::FETCH_REMAINING
		);
	}

	/**
	 * Fetch foregoing rows. Cursor will be in BEGINNING position after this operation.
	 * First returned row will be *previous* row.
	 * @return array[] foregoing rows
	 */
	public function fetchForegoing() : array
	{
		return $this->cursor->fetchRange(
			ICursor::FETCH_FOREGOING
		);
	}

	/**
	 * @param array $rows
	 * @return array|NULL
	 * @throws CursorException
	 */
	private function extractOneRowResult(array $rows)
	{
		$count = count($rows);
		if ($count === 1) {
			return reset($rows); // first element
		} elseif ($count === 0) {
			return NULL;
		}
		throw CursorException::unexpectedRowCountReturned($count);
	}

	/**
	 * @param array $row
	 * @return mixed The extracted value
	 */
	private function extractFirstColumnFromRow(array $row)
	{
		return reset($row);
	}



	// The ICursor proxy:

	/** {@inheritdoc} */
	public function getName() : string
	{
		return $this->cursor->getName();
	}

	/** {@inheritdoc} */
	public function isOnRecord() : bool
	{
		return $this->cursor->isOnRecord();
	}

	/** {@inheritdoc} */
	public function moveTo(int $index)
	{
		$this->cursor->moveTo($index);
	}

	/** {@inheritdoc} */
	public function moveFromEndTo(int $index)
	{
		return $this->cursor->moveFromEndTo($index);
	}

	/** {@inheritdoc} */
	public function moveBy(int $offset)
	{
		$this->cursor->moveBy($offset);
	}

	/** {@inheritdoc} */
	public function scroll(int $offset) : int
	{
		return $this->cursor->scroll($offset);
	}

	/** {@inheritdoc} */
	public function fetchRange(int $offset) : array
	{
		return $this->cursor->fetchRange($offset);
	}

	/** {@inheritdoc} */
	public function fetchOneAt(int $index)
	{
		return $this->cursor->fetchOneAt($index);
	}

	/** {@inheritdoc} */
	public function fetchOneBy(int $offset)
	{
		return $this->cursor->fetchOneBy($offset);
	}
}