<?php declare(strict_types = 1);
/**
 * This file is part of mappi/cursor.
 */

namespace Grifart\Mappi\Cursor;

class TrackedCursor implements ICursor
{
	/** @var Position */
	private $position;

	/** @var ICursor */
	private $cursor;

	/**
	 * Tip: if you are not sure that cursor will be in initial state, call
	 * ->moveToBeginning() after initialization.
	 * @param ICursor  $cursor Cursor in initial state (index=0)
	 * @param Position $initialPosition
	 */
	public function __construct(ICursor $cursor, Position $initialPosition)
	{
		$this->cursor = $cursor;
		$this->position = $initialPosition;
	}

	/**
	 * @return Position
	 */
	public function getPosition()
	{
		return clone $this->position;
	}

	public function getName() : string
	{
		return $this->cursor->getName();
	}

	public function isOnRecord() : bool
	{
		return $this->cursor->isOnRecord();
	}

	public function moveTo(int $index)
	{
		$this->cursor->moveTo($index);
		if ($this->cursor->isOnRecord()) {
			$this->position = Position::fromLeft($index);
			return;
		}

		if ($index === 0) {
			$this->position = Position::fromLeft(0); // BEGINNING
			return;
		}
		if ($index > 0 ) {
			$this->position = Position::fromRight(0); // END
			return;
		}
		throw CursorException::cursorPosition_unknownError();
	}

	public function moveFromEndTo(int $index)
	{
		$this->cursor->moveFromEndTo($index);
		if ($this->cursor->isOnRecord()) {
			$this->position = Position::fromRight($index);
			return;
		}

		if ($index === 0) {
			$this->position = Position::fromRight(0); // END
			return;
		}
		if ($index > 0) {
			$this->position = Position::fromLeft(0); // BEGINNING
			return;
		}
		throw CursorException::cursorPosition_unknownError();
	}

	public function moveBy(int $offset)
	{
		$this->cursor->moveBy($offset);
		if ($this->cursor->isOnRecord()) {
			$this->position = Position::relativeTo($this->position, $offset);
			return;
		}
		if ($offset < 0) {
			$this->position = Position::fromLeft(0); // BEGINNING
			return;
		}
		if ($offset > 0) {
			$this->position = Position::fromRight(0); // END
			return;
		}
		if ($offset === 0) {
			return; // already on one of ends
		}
		throw CursorException::cursorPosition_unknownError();
	}

	public function scroll(int $offset) : int
	{
		$wasOnRecord = $this->isOnRecord();
		$rowsRead = $cursorDelta = $this->cursor->scroll($offset);
		$forward = $offset >= 0;

		// scroll over the whole data-set
		$scrolledOverWholeSet = $rowsRead > 0 && $wasOnRecord === FALSE && !$this->isOnRecord();
		$wasOnRecordAndHitEnd = $wasOnRecord === TRUE && !$this->isOnRecord();
		if($scrolledOverWholeSet || $wasOnRecordAndHitEnd) {
			$cursorDelta++;
		}

		$this->position = Position::relativeTo(
			$this->position,
			($forward ? 1 : -1) * $cursorDelta
		);
		return $rowsRead;
	}

	public function fetchRange(int $offset): array
	{
		$result = $this->cursor->fetchRange($offset);
		if($offset === 0) { // zero does not move cursor
			return $result;
		}

		$forward = $offset > 0;
		$count = count($result);
		$reachedEnd = $count < abs($offset);
		if(!$reachedEnd) {
			$this->position = Position::relativeTo(
				$this->position,
				$count * ($forward ? 1 : -1)
			);
			return $result;
		}

		if($forward) {
			$this->position = Position::fromRight(0); // right END
			return $result;
		} else {
			$this->position = Position::fromLeft(0); // left END
			return $result;
		}
	}

	public function fetchOneAt(int $index)
	{ // todo: content tests
		$row = $this->cursor->fetchOneAt($index);
		if ($row !== NULL) {
			if($index < 0) {
				$this->position = Position::fromRight(abs($index));
			} else /* >= 0 */ {
				$this->position = Position::fromLeft(abs($index));
			}
			return $row;
		}

		if ($index === 0) {
			$this->position = Position::fromLeft(0); // BEGINNING
			return NULL;
		}

		if ($index < 0) { // went from right to left -> hit beginning
			$this->position = Position::fromLeft(0);
			return NULL;
		}
		if ($index > 0) { // went left->right -> hit end
			$this->position = Position::fromRight(0);
			return NULL;
		}

		throw CursorException::cursorPosition_unknownError();
	}

	public function fetchOneBy(int $offset)
	{ // todo: content tests
		$row = $this->cursor->fetchOneBy($offset);
		if ($row !== NULL) {
			$this->position = $this->position = Position::relativeTo($this->position, $offset);
			return $row;
		}

		if ($offset < 0) {
			$this->position = Position::fromLeft(0); // BEGINNING
			return NULL;
		}
		if ($offset > 0) {
			$this->position = Position::fromRight(0); // END
			return NULL;
		}
		if ($offset === 0) {
			return NULL; // already on one of ends
		}

		throw CursorException::cursorPosition_unknownError();
	}

}