<?php declare(strict_types = 1);
/**
 * This file is part of mappi/cursor.
 */

namespace Grifart\Mappi\Tests\Cursor\Driver;
use Grifart\Mappi\Cursor\CursorException;
use Grifart\Mappi\Cursor\Driver\ICursorDriver;

/**
 * Array Cursor simulates scrolling cursor as implemented in PostgreSQL 9.5.
 *
 * For more information see ICursor.
 * @see ICursor
 * @package Grifart\Mappi\Cursor
 */
class ArrayCursorDriver implements ICursorDriver
{
	private $cursors = [];
	private $indexes = [];

	/**
	 * Creates test cursor with given name
	 * @param string $name
	 * @param int    $length
	 */
	public function createTestCursor(string $name, int $length)
	{
		if(isset($this->cursors[$name])) {
			throw new CursorException("Cursor with name $name already exists");
		}

		$this->cursors[$name] = $this->generateCursorData($length);

		$this->indexes[$name] = 0;
	}

	/**
	 * Closes test cursor
	 * @param string $name
	 * @return bool
	 */
	public function close(string $name) : bool
	{
		unset($this->cursors[$name]);
		unset($this->indexes[$name]);
		return TRUE;
	}

	/**
	 * Cursor data is table with one column with name "n" which contains current row index as value
	 * @param int $length length of values
	 * @return array
	 */
	private function generateCursorData(int $length): array
	{
		$data = [];
		$data[0] = null;
		$data[$length+1] = null;
		for($i = 1; $i <= $length; $i++) {
			$data[$i] = ["n" => $i];
		}
		return $data;
	}

	/**
	 * Returns END position index from LEFT
	 * @param string $name The cursor name
	 * @return int
	 */
	private function getMaxIndex(string $name) {
		return count($this->cursors[$name])-1;
	}

	/**
	 * Normalizes index to be between BEGINNING and END of cursor
	 * @param string $name The cursor name
	 * @param int    $index The index to normalize
	 */
	private function normalizeIndex(string $name, int &$index)
	{
		if($index < 0) {
			$index = 0;
		}
		$max = $this->getMaxIndex($name);
		if($index > $max) {
			$index = $max;
		}
	}

	/**
	 * Moves internal cursor pointer to position from BEGINNING
	 * @param string $name The cursor name
	 * @param int    $index The index from the BEGINNING; index is normalized before move
	 */
	private function pointerMoveToFromBeginning(string $name, int $index)
	{
		$this->normalizeIndex($name, $index);
		$this->indexes[$name] = $index;
	}

	/**
	 * Moves internal cursor pointer to position from the END
	 * @param string $name The cursor name
	 * @param int    $index The index from the ENG; 
	 */
	private function pointerMoveToFromEnd(string $name, int $index)
	{
		$indexFromLeft = $this->getMaxIndex($name) - $index;
		$this->normalizeIndex($name, $indexFromLeft);

		$this->moveFromBeginningTo($name, $indexFromLeft);
	}
	
	private function getCurrentIndexFromBeginning(string $name) : int
	{
		return $this->indexes[$name];
	}

	private function retrieveCurrentRow(string $name) {
		return $this->cursors[$name][$this->getCurrentIndexFromBeginning($name)];
	}

	public function moveFromBeginningTo(string $name, int $index) : bool
	{
		if ($index < 0) {
			throw new \InvalidArgumentException("Negative index not supported. Use moveFromEndTo() instead.");
		}
		$this->pointerMoveToFromBeginning($name, $index);
		return !!$this->retrieveCurrentRow($name);
	}

	public function moveFromEndTo(string $name, int $index) : bool
	{
		if ($index < 0) {
			throw new \InvalidArgumentException("Negative index not supported. Use moveTo() instead.");
		}
		$this->pointerMoveToFromEnd($name, $index);
		return !!$this->retrieveCurrentRow($name);
	}

	public function moveBy(string $name, int $offset) : bool
	{
		$i = $this->getCurrentIndexFromBeginning($name);
		$indexFromLeft = $i + $offset;
		$this->normalizeIndex($name, $indexFromLeft);
		$this->moveFromBeginningTo($name, $indexFromLeft);
		return !!$this->retrieveCurrentRow($name);
	}

	public function scroll(string $name, int $offset) : int
	{
		if($offset === 0) return 0;
		return count($this->fetchRange($name, $offset));
	}

	// ----------------- FETCH ---------------------

	public function fetchRange(string $name, int $offset) : array
	{
		$currentKey = $this->getCurrentIndexFromBeginning($name);
		if($offset === self::FETCH_FOREGOING) {
			$finalKey = PHP_INT_MIN;
		} elseif ($offset === self::FETCH_REMAINING) {
			$finalKey = PHP_INT_MAX;
		} else {
			$finalKey = $currentKey + $offset;
		}
		$this->normalizeIndex($name, $finalKey);

		$data = [];
		if($offset === 0) {
			$v = $this->retrieveCurrentRow($name);
			if($v !== NULL) {
				$data[] = $v;
			}
			return $data;
		}

		if($finalKey < $currentKey) { // backwards
			for($i = $currentKey-1; $i >= $finalKey; $i--) {
				$this->moveFromBeginningTo($name, $i);
				$v = $this->retrieveCurrentRow($name);
				if($v === NULL) {break;}
				$data[] = $v;
			}
		} else {
			for($i = $currentKey+1; $i <= $finalKey; $i++) {
				$this->moveFromBeginningTo($name, $i);
				$v = $this->retrieveCurrentRow($name);
				if($v === NULL) {break;}
				$data[] = $v;
			}
		}
		return $data;
	}

	public function fetchOneAt(string $name, int $index)
	{
		if($index < 0) {
			$this->moveFromEndTo($name, abs($index));
		} else {
			$this->moveFromBeginningTo($name, abs($index));
		}
		$data = $this->fetchRange($name, 0);
		if(count($data) === 0) {
			return NULL;
		}
		return $data[0];
	}

	public function fetchOneBy(string $name, int $offset)
	{
		$this->moveBy($name, $offset);
		$data = $this->fetchRange($name, 0);
		if(count($data) === 0) {
			return NULL;
		}
		return $data[0];
	}

}