<?php
/**
 * @testCase
 */

namespace Grifart\Mappi\Tests\Cursor;

use Grifart\Mappi\Cursor\ICursor;
use Grifart\Mappi\Tests\Store\BaseTest;
use Tester\Assert;

/**
 * Every implementation of ICursor must pass this test.
 *
 * This file is using Given-When-Then naming convention.
 * @link http://martinfowler.com/bliki/GivenWhenThen.html
 *
 * @package Grifart\Mappi\Tests\Cursor
 */
abstract class ICursorTest extends BaseTest
{
	/** @var ICursor */
	protected $uut;



	protected function helper_fetchNextSingle()
	{
		$rows = $this->uut->fetchRange(1);
		if(count($rows) === 0) return NULL;
		return $rows[0]["n"];
	}

	protected function helper_fetchCurrentSingle()
	{
		$rows = $this->uut->fetchRange(0);
		if(count($rows) === 0) return NULL;
		return $rows[0]["n"];
	}

	protected function helper_fetchPrevSingle()
	{
		$rows = $this->uut->fetchRange(-1);
		if(count($rows) === 0) return NULL;
		return $rows[0]["n"];
	}



	// initial position
	public function test_givenCursor_whenGetCurrentRow_thenGetNothing()
	{
		Assert::count(0, $this->uut->fetchRange(0)); // must fail because initial position is 0
		Assert::count(0, $this->uut->fetchRange(-1));
		Assert::count(1, $this->uut->fetchRange(1));
	}

	public function test_givenInitialPosition_whenFetch_thenGetFirstRow()
	{
		Assert::same(1, $this->helper_fetchNextSingle());
	}

	// fetch( x == 0 )
	public function test_givenSecondPosition_whenFetch0_thenGetCurrentRow()
	{
		// head:           .
		// index: 0  1  2  3  4
		// value:    1  2  3  4
		$this->uut->moveTo(3);
		Assert::same(3, $this->helper_fetchCurrentSingle());

		$result = $this->uut->fetchRange(0); // rows backwards + read current

		Assert::count(1, $result);
		Assert::same(3, current($result[0]));
	}

	// fetch( x > 0 )
	public function test_givenSecondPosition_whenFetchForward_thenGetRowsAfterCursor()
	{
		// head:           .
		// index: 0  1  2  3  4
		// value:    1  2  3  4
		$this->uut->moveTo(3);
		Assert::same(3, $this->helper_fetchCurrentSingle());

		// head:                 .
		// index: 0  1  2  3  4  5  6
		// value:    1  2  3  4  5  6
		$result = $this->uut->fetchRange(2); // rows backwards + read current
		// todo: split into fetch decorator for those special fetch* methods

		Assert::count(2, $result);
		Assert::same(4, current($result[0]));
		Assert::same(5, current($result[1]));
		Assert::same(5, $this->helper_fetchCurrentSingle());
	}

	// fetch( x < 0 )
	public function test_givenSecondPosition_whenFetchBackwards_thenGetRowsBeforeCursor()
	{
		// head:           .
		// index: 0  1  2  3  4
		// value:    1  2  3  4
		$this->uut->moveTo(3);
		Assert::same(3, $this->helper_fetchCurrentSingle());

		// head:     .
		// index: 0  1  2  3  4
		// value:    1  2  3  4
		$result = $this->uut->fetchRange(-2); // rows backwards + read current
		// todo: split into fetch decorator for those special fetch* methods

		Assert::count(2, $result);
		Assert::same(2, current($result[0]));
		Assert::same(1, current($result[1]));
		Assert::same(1, $this->helper_fetchCurrentSingle());
	}


	// moveToBeginning()
	public function test_givenLastPosition_whenMoveToTheBeginning_thenFetchingNextRowWillBeFirstRow()
	{
		// Arrange
		$this->uut->moveFromEndTo(1);

		// Act
		$this->uut->moveTo(0);

		//Assert
		Assert::same(1, $this->helper_fetchNextSingle());
	}


	// moveToFirst()
	public function test_givenLastPosition_whenMoveToTheFirst_thenFetchingCurrentWillBeFirstRow()
	{
		// Arrange
		$this->uut->moveFromEndTo(1);

		// Act
		$this->uut->moveTo(1);

		//Assert
		Assert::same(1, $this->helper_fetchCurrentSingle());
	}


	// moveToEnd()
	public function test_givenInitialPosition_whenMoveToEnd_thenFetchingOneBacwardsWillBeTheLastRow()
	{
		$this->uut->moveFromEndTo(0);

		Assert::null($this->helper_fetchCurrentSingle());

		$data = $this->helper_fetchPrevSingle();
		Assert::same(1000, $data);
	}


	// moveToLast()
	public function test_givenInitialPosition_whenMoveToLast_thenFetchingCurrentWillBeLastRow()
	{
		$this->uut->moveFromEndTo(1);
		Assert::same(1000, $this->helper_fetchCurrentSingle());

		Assert::null($this->helper_fetchNextSingle());
	}


	// moveBy()
	public function test_givenInitialPosition_whenMoveBy2_thenGetSecondValue()
	{
		$this->uut->moveBy(2);
		Assert::same(2, $this->helper_fetchCurrentSingle());
	}

	public function test_givenSecondPosition_whenMoveOneBack_thenGetFirstValue()
	{
		$this->uut->moveTo(2);
		$this->uut->moveBy(-1);
		Assert::same(1, $this->helper_fetchCurrentSingle());
		$this->uut->moveBy(-1);
		Assert::null($this->helper_fetchCurrentSingle());
	}
	
	// scroll()
	public function test_givenInitialPosition_whenMoveToEnd_thenGetTotalCount()
	{
		Assert::same(1000, $this->uut->scroll(ICursor::FETCH_REMAINING));
		Assert::null($this->helper_fetchCurrentSingle());
	}

	public function test_givenFifthPosition_whenMoveToEndByIncrements_thenGetRestOfCount()
	{
		$this->uut->moveTo(5);
		Assert::same(5, $this->uut->scroll(5));
		Assert::same(10, $this->helper_fetchCurrentSingle());

		Assert::same(989, $this->uut->scroll(989));
		Assert::same(999, $this->helper_fetchCurrentSingle());

		Assert::same(1, $this->uut->scroll(1));
		Assert::same(1000, $this->helper_fetchCurrentSingle());

		Assert::same(0, $this->uut->scroll(1));
		Assert::null($this->helper_fetchCurrentSingle());

		Assert::same(0, $this->uut->scroll(1));
	}

	public function test_givenFifthPosition_whenScrollZero_thenGetOne()
	{
		$this->uut->moveTo(5);
		Assert::same(0, $this->uut->scroll(0));
	}

	public function test_givenFifthPosition_whenMoveToBeginning_thenGetForegoingNumberOfRows()
	{
		$this->uut->moveTo(5);

		Assert::same(4, $this->uut->scroll(ICursor::FETCH_FOREGOING));
	}

	public function test_givenBeginningPosition_whenMoveToBeginning_thenGetZero()
	{
		Assert::same(0, $this->uut->scroll(ICursor::FETCH_FOREGOING));
	}


	// fetchOneAt()
	public function test_givenInitialPosition_whenFetchBeginning_thenGetNull()
	{
		$data = $this->uut->fetchOneAt(0);
		Assert::null($data);
	}

	public function test_givenInitialPosition_whenFetchEnd_thenGetNull()
	{
		$data = $this->uut->fetchOneAt(1001);
		Assert::null($data);
	}

	public function test_givenInitialPosition_whenFetchLast_thenGetLast()
	{
		$data = $this->uut->fetchOneAt(1000);
		Assert::same(1000, $data["n"]);
	}

	public function test_givenInitialPosition_whenFetch5_thenGetFive()
	{
		$data = $this->uut->fetchOneAt(5);
		Assert::same(5, $data["n"]);
	}

	public function test_givenNonInitialPosition_whenFetch5_thenGetFive()
	{
		$this->uut->moveTo(234); // wherever

		$data = $this->uut->fetchOneAt(5);
		Assert::same(5, $data["n"]);
	}

	public function test_givenInitialPosition_whenFetchMinus2_thenGetSecondFromRight()
	{
		$this->uut->moveTo(234); // wherever

		$data = $this->uut->fetchOneAt(-2);
		Assert::same(999, $data["n"]);
	}


	// fetchOneBy()
	public function test_givenInitialPosition_whenFetchCurrent_thenGetNull()
	{
		Assert::null($this->uut->fetchOneBy(0));
	}

	public function test_givenInitialPosition_whenFetchFirst_thenGetFirst()
	{
		$data = $this->uut->fetchOneBy(1);
		Assert::same(1, $data["n"]);
	}

	public function test_givenFirstPosition_whenFetchLeft_thenGetNull()
	{
		$this->uut->moveTo(1);
		Assert::null($this->uut->fetchOneBy(-1));
	}

	public function test_givenMiddlePosition_whenFetchNext_thenGetNextPostion()
	{
		$this->uut->moveTo(500);
		Assert::same(501, $this->uut->fetchOneBy(1)["n"]);
	}

	public function test_givenLastPosition_whenFetchNext_thenGetNull()
	{
		$this->uut->moveFromEndTo(1);
		Assert::null($this->uut->fetchOneBy(1));
	}

	public function test_givenLastPosition_whenFetchCurrent_thenGetLastRow()
	{
		$this->uut->moveFromEndTo(1);
		Assert::same(1000, $this->uut->fetchOneBy(0)["n"]);
	}

	public function test_givenLastPosition_whenFetchLeft_thenGetLeftRow()
	{
		$this->uut->moveFromEndTo(1);
		Assert::same(999, $this->uut->fetchOneBy(-1)["n"]);
	}

	// fetchRemaining()
	public function test_given5BeforeEndPosition_whenFetchRemaining_thenGetLastFile()
	{
		$this->uut->moveTo(995);

		$result = $this->uut->fetchRange(ICursor::FETCH_REMAINING);

		Assert::count(5, $result);
		Assert::same(996,  current($result[0]));
		Assert::same(997,  current($result[1]));
		Assert::same(998,  current($result[2]));
		Assert::same(999,  current($result[3]));
		Assert::same(1000, current($result[4]));

		Assert::null($this->helper_fetchCurrentSingle());
		Assert::equal(1000, $this->helper_fetchPrevSingle());

	}


	// fetchRemaining()
	public function test_given5AfterStart_whenFetchForegoing_thenGetFirstFive()
	{
		$this->uut->moveTo(6);

		$result = $this->uut->fetchRange(ICursor::FETCH_FOREGOING);

		Assert::count(5, $result);
		Assert::same(5, current($result[0]));
		Assert::same(4, current($result[1]));
		Assert::same(3, current($result[2]));
		Assert::same(2, current($result[3]));
		Assert::same(1, current($result[4]));

		Assert::null($this->helper_fetchCurrentSingle());
	}


	// edge cases:
	public function test_givenInitialPosition_whenMoveToLeft_thenGetError()
	{
		Assert::false($this->uut->isOnRecord()); // initial position

		// start: index === 0
		$this->uut->moveBy(-1);
		Assert::false($this->uut->isOnRecord());

		// index === 0
		$data = $this->uut->fetchRange(-1);
		Assert::count(0, $data);

		$firstValue = $this->helper_fetchNextSingle();
		Assert::equal(1, $firstValue);
		// index === 1
	}

	public function test_givenEndPosition_whenMoveToRight_thenStayInPlace()
	{
		$this->uut->moveFromEndTo(0);

		Assert::null($this->helper_fetchNextSingle());
		Assert::null($this->helper_fetchCurrentSingle());
		Assert::equal(1000, $this->helper_fetchPrevSingle());
	}


	public function test_giveSomePosition_whenMoveToAfterEnd_thenWillBeAtTheEnd()
	{
		$this->uut->moveTo(25); // some position

		$this->uut->moveTo(9999);

		Assert::null($this->helper_fetchCurrentSingle());
		Assert::equal(1000, $this->helper_fetchPrevSingle());
	}

	public function test_givenFirstPosition_whenMoveToLeft_thenWillBeAtBeginning()
	{
		// Arrange
		Assert::false($this->uut->isOnRecord());
		$this->uut->moveTo(1);
		Assert::true($this->uut->isOnRecord());

		// Act
		$this->uut->moveBy(-1);

		// Assert
		Assert::false($this->uut->isOnRecord());
		Assert::null($this->helper_fetchCurrentSingle());
		Assert::null($this->helper_fetchPrevSingle());

		Assert::equal(1, $this->helper_fetchNextSingle());
		Assert::true($this->uut->isOnRecord());
	}

	public function test_givenLastPosition_whenMoveToRight_thenReachTheEnd()
	{
		$this->uut->moveFromEndTo(1);
		Assert::true($this->uut->isOnRecord());

		$this->uut->moveBy(1);

		Assert::false($this->uut->isOnRecord());
		Assert::null($this->helper_fetchCurrentSingle());
		Assert::null($this->helper_fetchNextSingle());

		Assert::equal(1000, $this->helper_fetchPrevSingle());
		Assert::true($this->uut->isOnRecord());
	}

	public function test_giveSomePosition_whenMoveToBeginning_thenWillBeAtBeginning()
	{
		$this->uut->moveTo(42);
		$this->uut->moveTo(0);
		Assert::equal(1,$this->helper_fetchNextSingle());
	}
	

}
