<?php
/**
 * @testCase
 */

namespace Grifart\Mappi\Tests\Cursor;

use Grifart\Mappi\Cursor\CursorException;
use Grifart\Mappi\Cursor\ICursor;
use Grifart\Mappi\Cursor\SemanticCursor;
use Grifart\Mappi\Tests\Store\BaseTest;
use Mockery;
use Tester\Assert;

require_once __DIR__ . "/../bootstrap.php";
require_once __DIR__ . "/ICursorTest.php";

/**
 * Behavioral test for SemanticCursor using mocking.
 * @link http://docs.mockery.io/en/latest/
 *
 * @package Grifart\Mappi\Tests\Cursor
 */
class SemanticCursorTest extends BaseTest
{
	/** @var SemanticCursor */
	private $uut;

	/** @var ICursor|Mockery\Mock */
	private $mockedCursor;

	protected function setUp()
	{
		parent::setUp();

		/** @var ICursor|Mockery\Mock $cursor */
		$this->mockedCursor = Mockery::mock(ICursor::class);
		$this->uut = new SemanticCursor($this->mockedCursor);
	}

	// ---- proxy ----
	public function test_getName()
	{
		$this->mockedCursor->shouldReceive("getName")
			->once()->andReturn("test");

		Assert::equal("test", $this->uut->getName());
	}

	public function test_isOnRecord()
	{
		$this->mockedCursor->shouldReceive("isOnRecord")
			->once()->andReturn(true);

		Assert::equal(true, $this->uut->isOnRecord());
	}

	public function test_moveTo()
	{
		$this->mockedCursor->shouldReceive("moveTo")
			->with(5)->once();

		$this->uut->moveTo(5);
	}

	public function test_moveFromEndTo()
	{
		$this->mockedCursor->shouldReceive("moveFromEndTo")
			->with(0)->once();

		$this->uut->moveFromEndTo(0);
	}

	public function test_moveBy()
	{
		$this->mockedCursor->shouldReceive("moveBy")
			->with(-8)->once();

		$this->uut->moveBy(-8);
	}

	public function test_fetchRange()
	{
		$this->mockedCursor->shouldReceive("fetchRange")
			->with(8)->once()->andReturn(["val1"]);

		Assert::equal(["val1"], $this->uut->fetchRange(8));
	}

	public function test_fetchOneAt()
	{
		$this->mockedCursor->shouldReceive("fetchOneAt")
			->with(5)->once()->andReturn(["hi!"]);

		Assert::equal(["hi!"], $this->uut->fetchOneAt(5));
	}

	public function test_fetchOneBy()
	{
		$this->mockedCursor->shouldReceive("fetchOneBy")
			->with(5)->once()->andReturn(["hi!"]);

		Assert::equal(["hi!"], $this->uut->fetchOneBy(5));
	}

	// ---- the extension - move* ----

	public function test_moveToLast()
	{
		$this->mockedCursor->shouldReceive("moveFromEndTo")
			->with(1)->once();

		$this->uut->moveToLast();
	}

	public function test_moveToFirst()
	{
		$this->mockedCursor->shouldReceive("moveTo")
			->with(1)->once();

		$this->uut->moveToFirst();
	}

	public function test_moveToBeginning()
	{
		$this->mockedCursor->shouldReceive("moveTo")
			->with(0)->once();

		$this->uut->moveToBeginning();
	}

	public function test_moveToEnd()
	{
		$this->mockedCursor->shouldReceive("moveFromEndTo")
			->with(0)->once();

		$this->uut->moveToEnd();
	}

	// ---- the extension - scroll* ----

	public function test_scrollBeginning() {
		$this->mockedCursor->shouldReceive("scroll")
			->with(ICursor::FETCH_FOREGOING)->once()->andReturn(5);

		Assert::same(5, $this->uut->scrollToBeginning());
	}

	public function test_scrollEnd() {
		$this->mockedCursor->shouldReceive("scroll")
			->with(ICursor::FETCH_REMAINING)->once()->andReturn(5);

		Assert::same(5, $this->uut->scrollToEnd());
	}

	// ---- the extension - fetch* ----

	public function test_fetchNext()
	{
		$this->mockedCursor->shouldReceive("fetchRange")
			->with(1)->once()->andReturn([["next"]]);

		Assert::equal(["next"], $this->uut->fetchNext());
	}

	public function test_fetchNext_givenEmptyResult_whenFetchNext_thenGetNull()
	{
		$this->mockedCursor->shouldReceive("fetchRange")
			->with(1)->once()->andReturn([]);

		Assert::null($this->uut->fetchNext());
	}

	public function test_fetchNext_givenMoreRowsReturned_whenFetchNext_thenGetError()
	{
		$this->mockedCursor->shouldReceive("fetchRange")
			->with(1)->once()->andReturn([["row1"],["row2"]]);

		Assert::exception(function() {
			$this->uut->fetchNext();
		}, CursorException::class, "Underlying driver failed. Returned unexpected number (2) of rows.");
	}

	public function test_fetchCurrent()
	{
		$this->mockedCursor->shouldReceive("fetchRange")
			->with(0)->once()->andReturn([["current"]]);

		Assert::equal(["current"], $this->uut->fetchCurrent());
	}

	public function test_fetchPrevious()
	{
		$this->mockedCursor->shouldReceive("fetchRange")
			->with(-1)->once()->andReturn([["previous"]]);

		Assert::equal(["previous"], $this->uut->fetchPrevious());
	}

	// ---- the extension - fetch*Single ----

	public function test_fetchNextSingle()
	{
		$this->mockedCursor->shouldReceive("fetchRange")
			->with(1)->once()->andReturn([["next", "col2"]]);

		Assert::equal("next", $this->uut->fetchNextSingle());
	}

	public function test_fetchCurrentSingle()
	{
		$this->mockedCursor->shouldReceive("fetchRange")
			->with(0)->once()->andReturn([["current"]]);

		Assert::equal("current", $this->uut->fetchCurrentSingle());
	}

	public function test_fetchPreviousSingle()
	{
		$this->mockedCursor->shouldReceive("fetchRange")
			->with(-1)->once()->andReturn([["previous"]]);

		Assert::equal("previous", $this->uut->fetchPreviousSingle());
	}

	public function test_fetchRemaining()
	{
		$this->mockedCursor->shouldReceive("fetchRange")
			->with(ICursor::FETCH_REMAINING)->once()
			->andReturn([["next1"],["next2"]]);

		Assert::equal(
			[["next1"],["next2"]],
			$this->uut->fetchRemaining()
		);
	}

	public function test_fetchForegoing()
	{
		$this->mockedCursor->shouldReceive("fetchRange")
			->with(ICursor::FETCH_FOREGOING)->once()
			->andReturn([["prev1"],["prev2"]]);

		Assert::equal(
			[["prev1"],["prev2"]],
			$this->uut->fetchForegoing()
		);
	}



}

(new SemanticCursorTest())->run();
